local g = BittensGlobalTables
local u, oldU, oldUVersion = g.MakeTable("BittensUtilities", 5)
if u == nil then
	return
end

local _G = _G
local CreateFrame = CreateFrame
local SlashCmdList = SlashCmdList
local next = next
local pairs = pairs
local select = select
local type = type
local unpack = unpack
local wipe = wipe

-------------------------------------------------------------- Table Management
function u.GetFromTable(outerTable, ...)
	for i = 1, select("#", ...) do
		if outerTable == nil then
			return nil
		end
		local key = select(i, ...)
		outerTable = outerTable[key]
	end
	return outerTable
end

function u.GetOrMakeTable(outerTable, ...)
	for i = 1, select("#", ...) do
		local key = select(i, ...)
		if outerTable[key] == nil then
			outerTable[key] = {}
		end
		outerTable = outerTable[key]
	end
	return outerTable
end

function u.MagicallyIncrement(table, variable)
	local val = (table[variable] or 0) + 1
	table[variable] = val
	return val
end

function u.SkipOrUpgrade(table, feature, version)
	local vName = feature .. "Version"
	if table[vName] and table[vName] >= version then
		return "Skip"
	else
		table[vName] = version
	end
end

function u.ShallowCopy(source, dest)
	if dest then
		wipe(dest)
	else
		dest = { }
	end
	for k, v in pairs(source) do
		dest[k] = v
	end
	return dest
end

function u.Pairs(table)
	if not table then
		return u.NoOp
	end
	
	local k, v
	return function()
		k, v = next(table, k)
		return k, v
	end
end

function u.Keys(table)
	local iterator = u.Pairs(table)
	return function()
		local k, v = iterator()
		return k
	end
end

function u.Values(table)
	local iterator = u.Pairs(table)
	return function()
		local k, v = iterator()
		return v
	end
end

function u.PrintTable(table)
	if not table then
		print("table is nil")
		return
	end
	
	local empty = true
	for k, v in pairs(table) do
		print(k, v)
		empty = false
	end
	
	if empty then
		print("table is empty")
	end
end

---------------------------------------------------------------- Slash Commands
local slashHandlers, oldSlashHandlers, oldSlashHandlersVersion 
	= g.GetOrMakeTable("BittensSlashHandlers", 1)
local commandHandlers = u.GetOrMakeTable(slashHandlers, "CommandHandlers")
local subCommandHandlers = u.GetOrMakeTable(slashHandlers, "SubCommandHandlers")

local function getOrMakeSlashHandler(command)
	local handler = commandHandlers[command]
	if handler == nil then
		handler = function(userInput)
			local subCommand, rest = userInput:match("^(%S*)%s*(.-)$")
			if subCommand == nil 
				or subCommandHandlers[command][subCommand] == nil then
				
				if subCommandHandlers[command]["<no sub-command>"] then
					subCommandHandlers[command]["<no sub-command>"](userInput)
				else
					print("Possible sub-commands:")
					for sub, _ in pairs(subCommandHandlers[command]) do
						print("   ", sub)
					end
				end
			else
				subCommandHandlers[command][subCommand](rest)
			end
		end
		commandHandlers[command] = handler
	end
	return handler
end

-- "command" can actually be a table of commands, if you wish
function u.RegisterSlashCommand(command, subCommand, func)
	local allCommands
	if type(command) == "table" then
		allCommands = command
	else
		allCommands = { command }
	end
	local fullName = "BITTENS_UTILS_" .. allCommands[1]
	
	for i, command in pairs(allCommands) do
		_G["SLASH_" .. fullName .. i] = "/" .. command
	end
	
	SlashCmdList[fullName] = getOrMakeSlashHandler(command)
	
	u.GetOrMakeTable(subCommandHandlers, command)
		[subCommand or "<no sub-command>"] = func
end

------------------------------------------------------------------------ Events
local eventToCustom = { }
local customToEvent = { }
local customFilters = { }

local function registerCustomEvent(customName, eventName, filter)
	eventToCustom[eventName] = customName
	customToEvent[customName] = eventName
	customFilters[customName] = filter
end

registerCustomEvent(
	"MY_ADDON_LOADED", 
	"ADDON_LOADED", 
	function(addonName, loadedAddonName)
		return addonName == loadedAddonName
	end)
registerCustomEvent(
	"MY_SPECIALIZATION_CHANGED", 
	"PLAYER_SPECIALIZATION_CHANGED",
	function(addonName, unit)
		return unit == "player"
	end)

function u.RegisterEventHandler(handler, addonName)
	local frame = CreateFrame("frame")
	for event, _ in pairs(handler) do
		local translated = customToEvent[event]
		if translated then
			if not handler[translated] then
				frame:RegisterEvent(translated)
			end
		else
			frame:RegisterEvent(event)
		end
	end
	frame:SetScript("OnEvent", function(self, event, ...)
		u.Call(handler[event], ...)
		local custom = eventToCustom[event]
		if custom and customFilters[custom](addonName, ...) then
			u.Call(handler[custom], ...)
		end
	end)
end

---------------------------------------------------------------- Options Panels

-- Add the functions apply() and refresh() to the returned frame.  They, along
-- with widgets in the frame, should use/modify the values in frame.Settings.
-- If you wish changes in widgets to apply immediately, call frame.refresh() 
-- after changing frame.Settings.
--
-- An empty table represents default values.
function u.CreateOptionsPanel(name, template, parent)	
	local frame = CreateFrame("Frame", nil, nil, template)
	frame.name = name
	
	function frame.Initialize(savedSettingsTableName)
		frame.SavedSettings = u.GetOrMakeTable(_G, savedSettingsTableName)
		frame.Settings = u.ShallowCopy(frame.SavedSettings)
		frame.apply()
	end
	
	function frame.AddLabel(text, template, x, y)
		local label = frame:CreateFontString(nil, nil, template)
		label:SetPoint("TOPLEFT", x, y)
		label:SetText(text)
		return y - label:GetHeight()
	end
	
	-- definition = {
	--     Name = string,
	--     Text = string,
	--     Default = boolean,
	-- }
	function frame.AddCheckBox(definition, x, y)
		local checkName = name .. definition.Name .. "CheckBox"
		local check = CreateFrame(
			"CheckButton", 
			checkName, 
			frame, 
			"InterfaceOptionsCheckButtonTemplate")
		check:SetPoint("TOPLEFT", x, y)
		_G[checkName .. "Text"]:SetText(definition.Text)
		check:SetScript("OnClick", function()
			frame.Settings[definition.Name] = not not check:GetChecked()
			frame.apply()
		end)
		u.GetOrMakeTable(frame, "CheckBoxes")[definition] = check
		return y - check:GetHeight()
	end
	
	function frame.AddCheckBoxGroup(label, definitions, x, y)
		y = frame.AddLabel(label, "GameFontNormalLarge", x, y)
		for _, definition in ipairs(definitions) do
			y = frame.AddCheckBox(definition, x, y)
		end
		return y
	end
	
	function frame.IsCheckBoxSelected(definition)
		local checked = frame.Settings[definition.Name]
		return checked == nil and definition.Default or checked
	end
	
	function frame.default()
		wipe(frame.Settings)
		frame.apply()
	end
	
	function frame.refresh()
		for definition, checkBox in u.Pairs(frame.CheckBoxes) do
			checkBox:SetChecked(frame.IsCheckBoxSelected(definition))
		end
		u.Call(frame.CustomRefresh)
	end
	
	function frame.okay()
		u.ShallowCopy(frame.Settings, frame.SavedSettings)
		frame.apply()
	end
	
	function frame.cancel()
		u.ShallowCopy(frame.SavedSettings, frame.Settings)
		frame.apply()
	end
	
	InterfaceOptions_AddCategory(frame)
	return frame
end

--------------------------------------------------------------------- functions
function u.Call(func, ...)
	if func then
		return func(...)
	end
end

function u.NoOp() end

function u.WrapFunction(func, ...)
	local args = { ... }
	return function()
		func(unpack(args))
	end
end
